""" Model a Bitwarden vault """

from datetime import datetime, timezone
from typing import List, Optional
import json
import logging
import os
import re
import subprocess

BW_CMD: str = "bw"
CLIENTID_ENVVAR = "BW_CLIENTID"
BW_SESSION_ENVVAR: str = "BW_SESSION"
DATADIR_ENVVAR: str = "BITWARDENCLI_APPDATA_DIR"
UNAUTHENTICATED: str = "unauthenticated"
LOCKED: str = "locked"
UNLOCKED: str = "unlocked"


def keyring_label(*, account_label: str, vault_uri: str) -> str:
    """
    Produce the formatted string that we are using to label our session keys in
    the kernel keyring
    """
    return f"{account_label}@{vault_uri}"


class VaultLockedError(Exception):
    """Custom Exception raised when an action fails because the vault is locked"""

    def __init__(  # pylint: disable=keyword-arg-before-vararg
        self, msg="Vault locked", *args, **kwargs
    ):
        super().__init__(msg, *args, **kwargs)


class VaultStatusNotRecognized(Exception):
    """
    Custom Exception raised when the vault status state is not recognized
    """

    def __init__(  # pylint: disable=keyword-arg-before-vararg
        self, msg="Vault status not recognized", *args, **kwargs
    ):
        super().__init__(msg, *args, **kwargs)


class VaultUnauthenticatedError(Exception):
    """
    Custom Exception raised when an action fails because the vault is unauthenticated
    """

    def __init__(  # pylint: disable=keyword-arg-before-vararg
        self, msg="Vault unauthenticated", *args, **kwargs
    ):
        super().__init__(msg, *args, **kwargs)


class Vault:  # pylint: disable=too-many-instance-attributes
    """
    This models/works with the BitWarden vault

    Basic intended flow:

    ```
    v = vault.Vault()
    try:
        v.get_session_from_keyring()
        v.get_items()
    except vault.VaultLockedError:
        v.unlock()
    except vault.VaultUnauthenticatedError:
        v.login()
        v.unlock()
    ```
    """

    def __init__(
        self,
        vault_uri: str = "https://vault.bitwarden.com",
        *,
        appdata_path: str,
        client_id: str,
        account_label: str,
    ):
        self.appdata_path: str = appdata_path
        self.vault_uri: str = vault_uri
        self.account_label: str = account_label
        self.client_id: str = client_id
        self.synctime_filepath: str = os.path.join(appdata_path, "last_sync")

        # Add to rather than overwrite existing environ
        self.env: dict = {
            **os.environ,
            **{
                DATADIR_ENVVAR: self.appdata_path,
                CLIENTID_ENVVAR: self.client_id,
            },
        }
        logging.debug("Setting environ for Vault interaction: %s", self.env)

        self.keyring_label = keyring_label(
            account_label=self.account_label, vault_uri=self.vault_uri
        )
        self.session_key: str

    def _record_sync_time(self) -> None:
        """
        Save to file the time of the last sync (by this program) for this vault

        Does not actually query the time of the last sync from `bw` itself, as
        the operation is expensive and ultimately unecessary

        we just need to know how long ago *this program* did it
        """
        now: datetime = datetime.now(timezone.utc)
        logging.info(
            "Recording new last sync time to `%s` as: `%s`",
            self.appdata_path,
            now.isoformat(),
        )
        os.makedirs(self.appdata_path, exist_ok=True)
        with open(
            self.synctime_filepath, "w", encoding="utf-8"
        ) as fh:  # pylint: disable=invalid-name
            fh.write(str(now.isoformat()))

    def last_sync_time(self) -> datetime:
        """
        Get the time that this program last synchronized the vault as an
        *aware* datetime object
        """
        try:
            with open(
                self.synctime_filepath, "r", encoding="utf-8"
            ) as fh:  # pylint: disable=invalid-name
                lasttime = fh.readline().strip("\n")
            return datetime.fromisoformat(lasttime)
        except FileNotFoundError:
            return datetime.fromtimestamp(0, timezone.utc)

    def sync(self) -> None:
        """
        Sync the vault and save the time
        """
        cmd: List[str]
        cmd = [BW_CMD, "sync"]
        logging.info("Syncing vault for `%s`", self.account_label)
        logging.debug(cmd)
        try:
            subprocess.run(  # type:ignore
                cmd,
                capture_output=True,
                check=True,
                env={**self.env, **{BW_SESSION_ENVVAR: self.session_key}},
                text=True,
            )
            self._record_sync_time()
        except subprocess.CalledProcessError as err:
            if err.returncode == 1:
                logging.info(
                    (
                        "`bw sync` failed with return code `1`. Presuming that "
                        "login is required. STDERR: %s"
                    ),
                    err.stderr,
                )
                raise VaultUnauthenticatedError from err

    def _list_objects(self, *, objtype: str) -> List[dict]:
        """
        Given an object type, get the list of that object from `bw` and return as JSON
        """
        objtypes: List[str] = ["items", "folders"]
        try:
            assert objtype in objtypes
        except AssertionError as err:
            raise AssertionError(
                f"`objtype` for `bw list` must be one of `{objtypes}`"
            ) from err

        cmd: List[str]
        cmd = [BW_CMD, "--nointeraction", "list", objtype]
        logging.info("Requesting vault objects of type `%s`", objtype)
        logging.debug(cmd)
        try:
            res: subprocess.CompletedProcess = subprocess.run(  # type:ignore
                cmd,
                capture_output=True,
                check=True,
                env={**self.env, **{BW_SESSION_ENVVAR: self.session_key}},
                text=True,
            )
        except subprocess.CalledProcessError as err:
            status: str = self.status()
            if status == LOCKED:
                raise VaultLockedError from err
            if status == UNAUTHENTICATED:
                raise VaultUnauthenticatedError from err
            if status != UNLOCKED:
                raise VaultStatusNotRecognized from err
        return json.loads(res.stdout)

    def get_items(
        self,
        *,
        fmt: str,
        prepend_folders: bool = False,
        prepend_str: Optional[str] = None,
    ) -> dict:
        """
        Return the items from the `bw list` processed into a dict with the
        actual menu listings as keys, and the secrets as the value.

        The secret value might not be from the password field - tries to return
        `login.password`->`secureNote`->`note` fields in that order of
        preference if exists
        and in that order
        """
        folders: dict
        items: list = self._list_objects(objtype="items")
        if prepend_folders:
            # If we're dealing with folders at all, then insert a new key
            # `folder_name` into each item with its folder name
            folders = {
                # Make into a dict of {folderid:foldername}
                obj["id"]: obj["name"]
                for obj in self._list_objects(objtype="folders")
            }
            for item in items:
                item["folder_name"] = folders[item["folderId"]]
        processed_items: dict = {}
        for item in items:
            name_parts: List[str] = []
            if prepend_str:
                name_parts.append(prepend_str)
            if prepend_folders:
                name_parts.append(item["folder_name"])
            name_parts.append(item["name"])
            secret: str
            if item.get("login", {}).get("password"):
                secret = item["login"]["password"]
            else:
                secret = item.get("notes", "")
            itemusername: str = item.get("login", {}).get("username", "")
            itemid: str = item["id"]
            itemname: str = "/".join(name_parts)
            listing_key: str = fmt.format(
                itemname=itemname,
                itemusername=itemusername,
                account_label=self.account_label,
                itemid=itemid,
                secret=secret,
            )
            processed_items[listing_key] = secret
        return processed_items

    def get_session_from_keyring(self) -> None:
        """
        Get session key, logging in and creating it if necessary
        """
        cmd: List[str]
        cmd = ["keyctl", "search", "@s", "user", self.keyring_label]
        logging.info(
            "Checking kernel keyring for existing session key for `%s@%s`",
            self.account_label,
            self.vault_uri,
        )
        logging.debug(cmd)
        res: subprocess.CompletedProcess = subprocess.run(  # type:ignore
            cmd, check=False, capture_output=True, text=True
        )
        keyid: str
        if res.returncode == 0:
            # If session key found
            keyid = res.stdout.strip("\n")
            logging.info(
                "Session keys for `%s@%s` found in keyring with id %s",
                self.account_label,
                self.vault_uri,
                keyid,
            )
            # Get the actual key
            cmd = ["keyctl", "pipe", keyid]
            logging.info(
                "Requesting `bw` session key from kernel keyring by id %s",
                keyid,
            )
            logging.debug(cmd)
            res: subprocess.CompletedProcess = subprocess.run(  # type:ignore
                cmd, check=True, capture_output=True, text=True
            )
            self.session_key = res.stdout
        else:
            # Else not found, unlock the vault and save the session key
            status: str = self.status()
            if status == LOCKED:
                raise VaultLockedError
            if status == UNAUTHENTICATED:
                raise VaultUnauthenticatedError
            if status != UNLOCKED:
                raise VaultStatusNotRecognized

    def status(self) -> str:
        """
        Return the status reported by `bw` for this vault
        """
        cmd: List[str] = [BW_CMD, "status"]
        logging.info(
            "Fetching vault status with additional environ: `%s=%s`",
            DATADIR_ENVVAR,
            self.appdata_path,
        )
        logging.debug(cmd, self.env)
        res: subprocess.CompletedProcess = subprocess.run(
            cmd, check=True, env=self.env, capture_output=True
        )
        return json.loads(res.stdout)["status"]

    def login(self, *, passphrase: str) -> None:  # pylint: disable=unused-argument
        """
        Log into `bw` via API key.

        This step is required before `bw unlock`
        """
        cmd: List[str] = [BW_CMD, "login", "--apikey"]
        logging.info(
            "Logging into vault with`%s=%s` (and password via STDIN)",
            DATADIR_ENVVAR,
            self.appdata_path,
        )
        logging.debug(cmd, self.env)
        try:
            subprocess.run(
                cmd,
                check=True,
                env=self.env,
                capture_output=True,
                text=True,
                input=passphrase,
            )
        except subprocess.CalledProcessError as err:
            stderr = err.stderr
            raise RuntimeError(
                f"`bw login` failed with returncode `{err.returncode}` and the "
                f"following from STDERR: `{stderr}`",
            ) from err

    def _store_session_key(self, *, session_key: str) -> None:
        """Save a session key and put a copy in the kernel keyring"""
        logging.info(
            "Updating stored session key in the Vault object for `%s",
            self.account_label,
        )
        self.session_key = session_key

        cmd: List[str]
        cmd = ["keyctl", "padd", "user", self.keyring_label, "@s"]
        logging.info(
            "Placing `bw` session key into kernel keyring via STDIN",
        )
        logging.debug(cmd)
        subprocess.run(  # type:ignore
            cmd,
            capture_output=True,
            check=True,
            input=session_key,
            text=True,
        )

    def _bw_extract_session_key(self, *, bw_output: str) -> str:
        """Given the output from `bw login` or `bw unlock`, extract the session key"""
        logging.debug(
            "Attempting to extract session key from `bw` output: %s",
            bw_output.encode("unicode_escape"),
        )
        match: Optional[re.Match] = re.search(
            BW_SESSION_ENVVAR + r'="([^"]+)"', bw_output
        )
        try:
            return match[1]  # type:ignore
        except (IndexError, AttributeError) as err:
            raise RuntimeError(
                "Failed to extract session key from `bw` output. Cannot proceed."
            ) from err

    def unlock(self, *, passphrase: str) -> None:
        """
        Unlock vault
        """
        cmd: List[str] = [BW_CMD, "unlock"]
        logging.info(
            "Unlocking vault with password via STDIN",
        )
        logging.debug(cmd, self.env)
        try:
            res: subprocess.CompletedProcess = subprocess.run(
                cmd,
                check=True,
                env=self.env,
                capture_output=True,
                text=True,
                input=passphrase,
            )
        except subprocess.CalledProcessError as err:
            stderr = err.stderr.encode("unicode_escape")
            raise RuntimeError(
                f"Vault unlock failed with returncode `{err.returncode}` and STDERR: "
                f"`{stderr}`"
            ) from err
        self._store_session_key(
            session_key=self._bw_extract_session_key(bw_output=res.stdout)
        )
        self._record_sync_time()



#!/usr/bin/env python3
"""
Pop-up BitWarden desktop quick-search. Puts the selection into the clipboard.
Persists the vault login for the duration of the login session.
"""

import argparse
import datetime
import logging
import os
import shlex
import subprocess
import sys
import tkinter as tk
import tkinter.simpledialog
from typing import Dict, List, Optional

from . import vault


def prompt_passphrase(text=""):
    """
    Prompt for text with hidden characters
    """
    logging.debug("Prompting for GUI input: `%s`", text)
    tk.Tk().withdraw()
    passphrase = tkinter.simpledialog.askstring(
        "Input Passphrase",
        text,
        show="*",
    )
    try:
        assert passphrase is not None
    except AssertionError as err:
        raise AssertionError("Input cannot be empty") from err
    return passphrase


def parse_args(argv=None) -> argparse.Namespace:
    """Parse args"""

    usage_examples: str = """examples:

        # Description

        %(prog)s <args>

    """
    descr: str = """
        Pop-up BitWarden desktop quick-search. Puts the selection into the
        clipboard. Persists the vault login for the duration of user's session.
        Supports multiple accounts simultaneously.

            %(prog)s --account arbitrary_account_label ~/path/to/appdata \\
                https://vault.domain.tld <api_client_id>

          Note that the account labels can be any unique value, as they are
          used only internally and are not used in interactions with the actual
          vault.

        Configuring the pop-up menu:

            `%(prog)s` tries to default to appropriate menu commands for
            your compositor - `dmenu` on X and `wofi` on Wayland.

            If you wish to override this, it is preferred to do so via
            evironment variables:

                # A custom prompt
                export MENU_CMD='dmenu -i -p "My custom prompt" -l 7'

                # Or even use a different menu program
                export MENU_CMD='rofi -dmenu -p "" -i'

                # Wayland
                export MENU_CMD='wofi -dmenu -p "" -i'

        Configuring the Clipboard:

            `%(prog)s` tries to default to appropriate clipboard commands for
            your compositor - `xclip` on X and `wl-copy` on Wayland.

            If you wish to override this, it is preferred to do so via
            evironment variables:

                # Xclip
                export CLIPBOARD_CMD='xclip -selection clipboard -in -loops 1'

                # Xsel
                export CLIPBOARD_CMD='xsel --selectionTimeout 5000 --input --clipboard'

                # On wayland, wl-copy
                export CLIPBOARD_CMD='wl-copy --paste-once'
        Examples:

            Type output via `xdotool` instead of putting to clipboard

                CLIPBOARD_CMD='xdotool -' %(prog)s --account account_label \\
                    ~/path/to/appdata https://vault.domain.tld <api_client_key> \\
                    --output-format-string 'type {passphrase}'

        Security:

            Note: The default clipboard commands limit persistence. Overriding
            the clipboard commands and failing to limit persistence could
            present a security risk. Be careful.

        """
    parser = argparse.ArgumentParser(
        description=descr,
        epilog=usage_examples,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    parser.add_argument(
        "--menu-format-string",
        "-f",
        help=(
            "String for Python's `str.format()` which determines how items are listed "
            "in dmenu. The fields made available are: `name`, `account_label`, "
            "`itemid`, and `secret`"
        ),
        default="{itemname} [{itemusername}] [{itemid}]",
        type=str,
    )

    parser.add_argument(
        "--sync-interval",
        "-i",
        help=(
            "The BitWarden CLI does not automatically pull new information after "
            "login. This option allows the user to specify a maximum interval (in"
            "minutes) after which this tool will sync the local vault after running. A"
            "negative value prevents this application from syncing the vault."
        ),
        default=60,
        type=int,
    )

    parser.add_argument(
        "--print-by-id",
        help=(
            "If a specific item ID is given here, `%(prog)s` will not prompt for a "
            "selection or output to the clipboard, instead printing the given id "
            "directly to STDOUT."
        ),
        nargs=2,
        type=str,
    )

    parser.add_argument(
        "--prepend-folders",
        "-p",
        action=argparse.BooleanOptionalAction,
        help=(
            "If set, will prepend folder names to the items listed. Note: this will "
            "make the run take longer, as the BitWarden CLI must be called twice."
        ),
        default=False,
        type=bool,
    )

    parser.add_argument(
        "--prepend-account-label",
        "-l",
        action=argparse.BooleanOptionalAction,
        help=(
            "If set, will prepend the account label to the items listed. Useful if "
            "using multiple accounts."
        ),
        default=False,
        type=bool,
    )

    parser.add_argument(
        "--output-format-string",
        "-o",
        help=(
            "String for Python's `str.format()` which allows dressing up of the"
            "output to the clipboard. The only fields available is `passphrase`."
            "This can be used in combination with the `CLIPBOARD_CMD`"
            "environment variable to, for example, have this tool type the"
            "selection out automatically. See `%(prog)s --help` for more."
        ),
        default="{passphrase}",
        type=str,
    )

    parser.add_argument(
        "--include-session-key",
        "-k",
        action=argparse.BooleanOptionalAction,
        help=(
            "If set, this tool will include the active session key in the dmenu output"
        ),
        default=False,
        type=bool,
    )

    # Prepare default menu and clipboard commands for X/Wayland
    default_clipboard_cmd: str
    default_menu_command: str
    if "wayland" in os.environ.get("XDG_SESSION_TYPE", ""):
        default_clipboard_cmd = "wl-copy --paste-once"
        default_menu_command = "wofi -dmenu -p '' -i"
    else:
        default_clipboard_cmd = "xclip -selection clipboard -in -loops 1"
        default_menu_command = "dmenu -i -p BitWarden -l 7"
    parser.add_argument(
        "--clipboard-cmd",
        "-c",
        default=shlex.split(os.environ.get("CLIPBOARD_CMD", default_clipboard_cmd)),
        type=list,
    )
    parser.add_argument(
        "--menu-cmd",
        "-d",
        help=argparse.SUPPRESS,
        default=shlex.split(os.environ.get("MENU_CMD", default_menu_command)),
        type=str,
    )

    parser.add_argument(
        "--account",
        "-a",
        metavar=("ARBITRARY_LABEL", vault.DATADIR_ENVVAR, "VAULT_URI", "API_CLIENT_ID"),
        action="append",
        dest="accounts",
        help=(
            "The details necessary to log in to a vault with API key. "
            "`ARBITRARY_LABEL` is not used in any interaction with the vault"
            "(that is what the API keys are for), but only to differentiate "
            "between multiple accounts internally."
        ),
        nargs=4,
        required=True,
        type=str,
    )

    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        dest="verbosity",
        help="Set output verbosity on STDERR (-v=warning, -vv=debug)",
    )

    args = parser.parse_args(argv) if argv else parser.parse_args()

    if args.print_by_id:
        try:
            assert args.print_by_id[0] in [a[0] for a in args.accounts]
        except AssertionError as err:
            raise AssertionError(
                "The account label for `--print-by-id` must also be present in "
                "`--account`"
            ) from err

    if args.verbosity >= 2:
        log_level = logging.DEBUG
    elif args.verbosity >= 1:
        log_level = logging.INFO
    else:
        log_level = logging.WARNING

    logging.basicConfig(level=log_level)

    return args


def put_to_clipboard(*, clipboard_cmd: list, passphrase: str) -> None:
    """
    Given a command (like the default `xclip`), output the selection to it
    """
    try:
        logging.info("Outputting to clipboard command")
        logging.debug(clipboard_cmd)
        with subprocess.Popen(
            clipboard_cmd,
            stdin=subprocess.PIPE,
            text=True,
        ) as clipboard_process:
            # Work with xclip more directly, as `xclip` leaves its stream
            # open and Python's `subprocess.run()` will `wait()` on it forever
            clipboard_process.communicate(passphrase)
    except subprocess.CalledProcessError as err:
        raise RuntimeError("Clipboard placement failed.") from err


def prompt_selection(*, items: List[str], menu_cmd: str) -> str:
    """
    Call the menu command with our item list as input
    return the selected key
    """
    try:
        logging.info("Running external menu command")
        logging.debug(menu_cmd)
        res: subprocess.CompletedProcess = subprocess.run(
            menu_cmd,
            capture_output=True,
            check=True,
            text=True,
            input="\n".join(items),
        )
    except subprocess.CalledProcessError as err:
        raise RuntimeError("Menu prompt failed.") from err
    return res.stdout.strip("\n")


def sync_after_interval(*, uservault: vault.Vault, interval: int) -> None:
    """
    Sync the vault if the last sync time exceeds the limit
    """
    now: datetime.datetime = datetime.datetime.now(datetime.timezone.utc)
    syncdiff_minutes = (now.timestamp() - uservault.last_sync_time().timestamp()) / 60
    logging.debug(
        "Time since last sync `%s` minutes vs max interval `%s` minutes",
        syncdiff_minutes,
        interval,
    )
    if syncdiff_minutes > interval >= 0:
        uservault.sync()


def main(argv: Optional[list] = None) -> None:  # pylint: disable=too-many-locals
    """Main"""
    args: argparse.Namespace
    if argv:
        args = parse_args(argv)
    else:
        args = parse_args(sys.argv[1:])
    logging.debug("Argparse results: %s", args)

    items: Dict[str, str] = {}
    account: List[str]
    vaults: List[vault.Vault] = []
    for account in args.accounts:
        account_label: str
        datadir_path: str
        vault_uri: str
        client_id: str
        account_label, datadir_path, vault_uri, client_id = account

        uservault: vault.Vault = vault.Vault(
            appdata_path=datadir_path,
            vault_uri=vault_uri,
            account_label=account_label,
            client_id=client_id,
        )

        prepend_str: Optional[str] = (
            account_label if args.prepend_account_label else None
        )
        user_items: dict
        menu_format_string: str = (
            "{account_label}|{itemid}" if args.print_by_id else args.menu_format_string
        )
        try:
            uservault.get_session_from_keyring()
            user_items = uservault.get_items(
                prepend_folders=args.prepend_folders,
                fmt=menu_format_string,
                prepend_str=prepend_str,
            )
        except vault.VaultLockedError:
            logging.debug("Vault locked. Prompting for credentials to unlock.")
            passphrase = prompt_passphrase(
                text=f"Enter passphrase for user `{account_label}` @ `{vault_uri}`"
            )
            uservault.unlock(passphrase=passphrase)
            user_items = uservault.get_items(
                prepend_folders=args.prepend_folders,
                fmt=menu_format_string,
                prepend_str=prepend_str,
            )
        except vault.VaultUnauthenticatedError:
            logging.debug("Vault unauthenticated. Prompting for credentials to login.")
            uservault.login(
                passphrase=prompt_passphrase(
                    text=(
                        f"Enter API client secret for user `{account_label}` @ "
                        f"`{vault_uri}`"
                    )
                )
            )
            passphrase = prompt_passphrase(
                text=f"Enter passphrase for user `{account_label}` @ `{vault_uri}`"
            )
            uservault.unlock(passphrase=passphrase)
            user_items = uservault.get_items(
                prepend_folders=args.prepend_folders,
                fmt=menu_format_string,
                prepend_str=prepend_str,
            )
        if args.include_session_key:
            # If the user would like the seesion keys for eahc account included
            # in the selection menu
            user_items[f"BW_SESSION for {account_label}"] = uservault.session_key
        items = items | user_items
        vaults.append(uservault)

    if args.print_by_id:
        try:
            print(items["|".join(args.print_by_id)])
        except KeyError as err:
            raise KeyError(
                "The account label and item id given to `--print-by-id` were not found"
            ) from err
    else:
        selection: str = items[
            prompt_selection(menu_cmd=args.menu_cmd, items=list(items))
        ]
        output: str = args.output_format_string.format(passphrase=selection)
        put_to_clipboard(clipboard_cmd=args.clipboard_cmd, passphrase=output)

        # Finally ensure we are in sync, but after user interaction is complete
        # to avoid making the user wait
        for uservault in vaults:
            sync_after_interval(uservault=uservault, interval=args.sync_interval)


if __name__ == "__main__":
    main()



"""
Test the full run of the program with one user, end to end, with multiple
scenarios
"""

import atexit
import importlib
import os
import shutil
import sys
import tempfile

from pathlib import Path
from typing import List
from datetime import datetime, timezone

import jinja2 as jinja
import pytest  # type:ignore
import testpath  # type:ignore
import bw_menu.__main__ as program
from bw_menu import vault

UNAUTHENTICATED: str = "unauthenticated"
LOCKED: str = "locked"
UNLOCKED: str = "unlocked"


USER1: str = "user1@domain.tld"
USER1_APPDATA_DIR: str = tempfile.mkdtemp(prefix="USER1")
VAULT_URI: str = "https://vault.bitwarden.com"
USER1_CLIENT_ID: str = "user.001111111q111vvv"
PASSWORD: str = "PASSWORD"
BW_STATUS_FILE_PATH: str
_, BW_STATUS_FILE_PATH = tempfile.mkstemp()

with open("tests/data/list_items.json", mode="r", encoding="utf-8") as jsonf:
    BW_OUTPUT_ITEMS: str = jsonf.read()
with open("tests/data/list_folders.json", mode="r", encoding="utf-8") as jsonf:
    BW_OUTPUT_FOLDERS: str = jsonf.read()


def cleanup_temp_files():
    """clean up"""
    shutil.rmtree(USER1_APPDATA_DIR)
    Path.unlink(Path(BW_STATUS_FILE_PATH), missing_ok=True)


atexit.register(cleanup_temp_files)


with open("tests/templates/bw.j2", mode="r", encoding="utf-8") as fh:
    bw_template = jinja.Template(fh.read())
    BW_SCRIPT = bw_template.render(
        bw_status_file_path=BW_STATUS_FILE_PATH,
        bw_output_items=BW_OUTPUT_ITEMS,
        bw_output_folders=BW_OUTPUT_FOLDERS,
    )

KEYCTL_KEY_ID: str = "000000000"
KEYCTL_KEY_CONTENTS: str = PASSWORD
with open("tests/templates/keyctl.j2", mode="r", encoding="utf-8") as fh:
    bw_template = jinja.Template(fh.read())
    KEYCTL_SCRIPT = bw_template.render(
        keyctl_key_id=KEYCTL_KEY_ID,
    )

MENU_SCRIPT: str = """
#!/usr/bin/env python3
import sys
print(sys.stdin.readline())
"""

KEYCTL_CMDS: dict = {
    "search": [
        "search",
        "@s",
        "user",
        vault.keyring_label(account_label=USER1, vault_uri=VAULT_URI),
    ],
    "pipe": ["pipe", KEYCTL_KEY_ID],
    "padd": [
        "padd",
        "user",
        vault.keyring_label(account_label=USER1, vault_uri=VAULT_URI),
        "@s",
    ],
}
BW_CMDS: dict = {
    "folders": ["--nointeraction", "list", "folders"],
    "items": ["--nointeraction", "list", "items"],
    "login": ["login", "--apikey"],
    "status": ["status"],
    "sync": ["sync"],
    "unlock": ["unlock"],
}

DMENU_CMD = ["-i", "-p", "BitWarden", "-l", "7"]
CLIPBD_CMD = ["-selection", "clipboard", "-in", "-loops", "1"]


@pytest.mark.parametrize(
    "test_input,expected",
    [
        (
            # A totally new user with an unauthenticated vault
            # Keyring shows old keys, requiring updating
            [
                "--account",
                USER1,
                USER1_APPDATA_DIR,
                VAULT_URI,
                USER1_CLIENT_ID,
            ],
            {
                "returncode": 0,
                "output": "mypassword",
                "vault_start": UNAUTHENTICATED,
                "bw_args": [
                    BW_CMDS["items"],
                    BW_CMDS["status"],
                    BW_CMDS["login"],
                    BW_CMDS["unlock"],
                    BW_CMDS["items"],
                ],
                "keyctl_args": [
                    KEYCTL_CMDS["search"],
                    KEYCTL_CMDS["pipe"],
                    KEYCTL_CMDS["padd"],
                ],
                "dmenu_args": [DMENU_CMD],
                "clipbd_args": [CLIPBD_CMD],
            },
        ),
        (
            # Vault authenticated but not unlocked
            # Keyring shows old keys, requiring updating
            [
                "--account",
                USER1,
                USER1_APPDATA_DIR,
                VAULT_URI,
                USER1_CLIENT_ID,
            ],
            {
                "returncode": 0,
                "output": "mypassword",
                "vault_start": LOCKED,
                "bw_args": [
                    BW_CMDS["items"],
                    BW_CMDS["status"],
                    BW_CMDS["unlock"],
                    BW_CMDS["items"],
                ],
                "keyctl_args": [
                    KEYCTL_CMDS["search"],
                    KEYCTL_CMDS["pipe"],
                    KEYCTL_CMDS["padd"],
                ],
                "dmenu_args": [DMENU_CMD],
                "clipbd_args": [CLIPBD_CMD],
            },
        ),
        (
            # Vault already unlocked with valid keys in keyring
            [
                "--account",
                USER1,
                USER1_APPDATA_DIR,
                VAULT_URI,
                USER1_CLIENT_ID,
            ],
            {
                "returncode": 0,
                "output": "mypassword",
                "vault_start": UNLOCKED,
                "bw_args": [
                    BW_CMDS["items"],
                ],
                "keyctl_args": [
                    KEYCTL_CMDS["search"],
                    KEYCTL_CMDS["pipe"],
                ],
                "dmenu_args": [DMENU_CMD],
                "clipbd_args": [CLIPBD_CMD],
            },
        ),
        (
            # Vault already unlocked with valid keys in keyring. User requests
            # that folder names are prepended to the listing
            [
                "--account",
                USER1,
                USER1_APPDATA_DIR,
                VAULT_URI,
                USER1_CLIENT_ID,
                "--prepend-folders",
            ],
            {
                "returncode": 0,
                "output": "mypassword",
                "vault_start": UNLOCKED,
                "bw_args": [
                    BW_CMDS["items"],
                    BW_CMDS["folders"],
                ],
                "keyctl_args": [
                    KEYCTL_CMDS["search"],
                    KEYCTL_CMDS["pipe"],
                ],
                "dmenu_args": [DMENU_CMD],
                "clipbd_args": [CLIPBD_CMD],
            },
        ),
        (
            # Vault already unlocked with valid keys in keyring, and a last
            # sync time that is over the default time
            [
                "--account",
                USER1,
                USER1_APPDATA_DIR,
                VAULT_URI,
                USER1_CLIENT_ID,
            ],
            {
                "returncode": 0,
                "output": "mypassword",
                "vault_start": UNLOCKED,
                "last_sync_time": datetime.fromtimestamp(0, timezone.utc),
                "bw_args": [
                    BW_CMDS["items"],
                    BW_CMDS["sync"],
                ],
                "keyctl_args": [
                    KEYCTL_CMDS["search"],
                    KEYCTL_CMDS["pipe"],
                ],
                "dmenu_args": [DMENU_CMD],
                "clipbd_args": [CLIPBD_CMD],
            },
        ),
        (
            # Vault already unlocked with valid keys in keyring. Should print a
            # specific ID to stdout.
            [
                "--account",
                USER1,
                USER1_APPDATA_DIR,
                VAULT_URI,
                USER1_CLIENT_ID,
                "--print-by-id",
                USER1,
                "item_id",
            ],
            {
                "returncode": 0,
                "stdout": "PASSWORD",
                "vault_start": UNLOCKED,
                "bw_args": [
                    BW_CMDS["items"],
                ],
                "keyctl_args": [
                    KEYCTL_CMDS["search"],
                    KEYCTL_CMDS["pipe"],
                ],
                "dmenu_args": [],
                "clipbd_args": [],
            },
        ),
    ],
)
# pylint: disable=unused-argument
# pylint: disable=redefined-outer-name
def test_end_to_end(test_input, expected, capsys) -> None:
    """Test"""
    args: List[str] = [str(x) for x in test_input]
    with testpath.MockCommand("bw", python=BW_SCRIPT,) as bwcli, testpath.MockCommand(
        "keyctl",
        python=KEYCTL_SCRIPT,
    ) as keyctl, testpath.MockCommand(
        "dmenu",
        python=MENU_SCRIPT,
    ) as dmenu, testpath.MockCommand.fixed_output(
        "xclip", ""
    ) as clipbd:
        # Mock Tkinter to just pass back a string as if the user had filled it
        # out
        sys.modules["tkinter"] = importlib.import_module("tkinter")
        with open(BW_STATUS_FILE_PATH, mode="w", encoding="utf-8") as bw_status_file:
            bw_status_file.write(expected["vault_start"])
        if "last_sync_time" in expected:
            with open(
                os.path.join(USER1_APPDATA_DIR, "last_sync"), mode="w", encoding="utf-8"
            ) as last_sync_file:
                last_sync_file.write(str(expected["last_sync_time"]))
        program.main(argv=args)

    if "stdout" in expected:
        assert capsys.readouterr().out.strip("\n") == expected["stdout"]

    for mock_cmd, all_expected_args in [
        (bwcli, expected["bw_args"]),
        (keyctl, expected["keyctl_args"]),
        (dmenu, expected["dmenu_args"]),
        [clipbd, expected["clipbd_args"]],
    ]:
        # Go through each mocked external command and ensure that they were
        # all called with the right arguments and in the right order
        cmds_run: List[str] = [a["argv"][1:] for a in mock_cmd.get_calls()]
        # print(cmds_run)
        # print(all_expected_args)
        assert len(cmds_run) == len(all_expected_args)
        for called_args, expected_args in zip(cmds_run, all_expected_args):
            assert " ".join(called_args) == " ".join(expected_args), (
                "The commands that we expected to be called must all have run, and "
                "in the same order - ",
            )


# pylint: enable=unused-argument
# pylint: enable=redefined-outer-name
